# -*- coding: utf-8 -*-
"""Simple script for running a PhonopyCalculation."""
from aiida import load_profile, orm
from aiida.engine import submit
import phonopy

from aiida_phonopy.calculations.phonopy import PhonopyCalculation
from aiida_phonopy.data import PhonopyData, PreProcessData

load_profile()


def main():
    """Run a simple Phonopy post-processing using PhonopyCalculation."""
    ph = phonopy.load('./phonopy.yaml', force_sets_filename='./FORCE_SETS')

    preprocess_data = PreProcessData(phonopy_atoms=ph.unitcell, supercell_matrix=[3, 3, 3])
    preprocess_data.set_displacements_from_dataset(dataset=ph.dataset)

    phonopy_data = PhonopyData(preprocess_data=preprocess_data)
    phonopy_data.set_forces(sets_of_forces=ph.forces)

    code = orm.load_code('phonopy@localhost')
    parameters = {'IRREPS': [0, 0, 0], 'FC_SYMMETRY': True, 'WRITE_FORCE_CONSTANTS': True}

    inputs = {
        'code': code,
        'parameters': orm.Dict(parameters),
        'phonopy_data': phonopy_data,
        'metadata': {
            'options': {
                'resources': {
                    'num_machines': 1
                },
            }
        }
    }

    submit(PhonopyCalculation, **inputs)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Functions for linking PhonopyAtoms and StructureData."""

__all__ = ('phonopy_atoms_to_structure', 'phonopy_atoms_from_structure', 'if_to_map')


def phonopy_atoms_to_structure(cell, mapping=None):
    """Return a StructureData from a PhonopyAtoms instance.

    :param cell: a PhonopyAtoms instance
    :param mapping: a number to kinds and symbols map, defaults to None
    """
    from aiida import orm

    if mapping:
        numbers_to_kinds, numbers_to_symbols = mapping
        symbols = []
        positions = []
        names = []
        for number in cell.numbers:
            try:
                names.append(numbers_to_kinds[number])
                symbols.append(numbers_to_symbols[number])
            except KeyError:
                names.append(numbers_to_kinds[str(number)])
                symbols.append(numbers_to_symbols[str(number)])
    else:
        names = cell.symbols
        symbols = cell.symbols

    positions = cell.positions
    masses = cell.masses

    structure = orm.StructureData(cell=cell.cell)
    for position, symbol, name, mass in zip(positions, symbols, names, masses):
        structure.append_atom(position=position, symbols=symbol, name=name, mass=mass)

    return structure


def phonopy_atoms_from_structure(structure):
    """Return a tuple containg the PhonopyAtoms and the mapping from a StructureData.

    :param structure: StructureData instance
    :return: it returns a PhonopyAtoms istance and the mapping
    """
    from phonopy.structure.cells import PhonopyAtoms

    to_map = if_to_map(structure)

    if not to_map:  # when kind_names=symbols are used in the structure
        cell = PhonopyAtoms(
            symbols=[site.kind_name for site in structure.sites],
            positions=[site.position for site in structure.sites],
            cell=structure.cell,
        )
        return (cell, None)
    # when using custom kind_names (!=symbols) in the structure
    sites = structure.sites
    symbols = []  # chemical symbols
    positions = []
    masses = []
    names = []  # actual kind names
    numbers = []

    kind_names = structure.get_kind_names()
    # The numbers start from 1, i.e. from the hydrogen element.
    kinds_to_numbers = {kind_names[i]: i + 1 for i in range(len(kind_names))}

    for site in sites:
        kind = structure.get_kind(site.kind_name)
        symbols.append(kind.symbol)
        masses.append(kind.mass)
        names.append(kind.name)
        positions.append(site.position)
        numbers.append(kinds_to_numbers[kind.name])

    numbers_to_symbols = {numbers[i]: symbols[i] for i in range(len(sites))}
    # I build a PhonopyAtoms instance using the numbers instead of the symbols.
    # This will change the symbols which will be automatically stored in the instance.
    # We keep track of this 'issue' by returning `mapping` used
    # afterwards to remap the kind_names to the atoms.
    cell = PhonopyAtoms(
        numbers=numbers,
        positions=positions,
        masses=masses,
        cell=structure.cell,
    )

    # Also here we start from 1.
    numbers_to_kinds = {i + 1: kind_names[i] for i in range(len(kind_names))}

    return (cell, [numbers_to_kinds, numbers_to_symbols])


def if_to_map(structure):
    """Return a bool according whether kind names and symbols are the same.

    :param structure: StructureData instance
    :return: True if kind names different from symbols are used, False otherwise
    """
    check_kinds = [True for kind in structure.kinds if kind.name != kind.symbol]
    return bool(check_kinds)



# -*- coding: utf-8 -*-
"""CalcJob for phonopy post-processing."""

from aiida import orm
from aiida.common import InputValidationError, datastructures
from aiida.engine import CalcJob

from aiida_phonopy.data import ForceConstantsData, PhonopyData
from aiida_phonopy.utils.mapping import _lowercase_dict, _uppercase_dict


def get_default_metadata_options():
    """Get a default metadata option for Phonopy calculation."""


class PhonopyCalculation(CalcJob):
    """Base `CalcJob` implementation for Phonopy post-processing."""

    # Mapping keys for parsers, giving respectively output file name and output node name.
    _OUTPUTS = {
        # format: .hdf5
        'fc': ('force_constants.hdf5', 'output_force_constants'),
        'band': ('band.hdf5', 'phonon_bands'),
        'qpoints': ('qpoints.hdf5', 'qpoints'),
        'mesh': ('mesh.hdf5', 'qpoints_mesh'),
        # format: .yaml
        'irreps': ('irreps.yaml', 'irreducible_representations'),
        'tprop': ('thermal_properties.yaml', 'thermal_properties'),
        'tdisp': ('thermal_displacements.yaml', 'thermal_displacements'),
        'tdispmat': ('thermal_displacement_matrices.yaml', 'thermal_displacement_matrices'),
        'mod': ('modulation.yaml', 'modulation'),
        # format: .dat
        'dos': ('total_dos.dat', 'total_phonon_dos'),
        'pdos': ('projected_dos.dat', 'projected_phonon_dos')
    }

    _INPUT_FORCE_CONSTANTS = 'force_constants.hdf5'

    _DEFAULT_INPUT_FILE = 'aiida.in'
    _DEFAULT_OUTPUT_FILE = 'aiida.out'
    _DEFAULT_PHONOPY_FILE = 'phonopy.yaml'

    _INPUT_SUBFOLDER = './'
    _OUTPUT_SUBFOLDER = './'

    # Available tags - further selection has to be made
    _AVAILABLE_TAGS = {
        # Basic tags
        'PRIMITIVE_AXES': [list],  # delete?
        'PRIMITIVE_AXIS': [list],
        'EIGENVECTORS': [bool],
        # Band structure tags
        'BAND': [str, list],
        'BAND_PATHS': [str, list],
        'BAND_POINTS': [float],
        'BAND_LABELS': [list],
        'BAND_CONNECTION': [bool],
        'BAND_INDICES': [list],
        # Mehs sampling tags
        'MESH': [list],
        'MP': [int, float],
        'MESH_NUMBERS': [int, float],  # ?
        'MP_SHIFT': [list],  # ?
        'GAMMA_CENTER': [bool],
        'WRITE_MESH': [bool],
        # Phonon density of states (DOS) tags
        'DOS': [bool],
        'DOS_RANGE': [str, list],
        'FMIN': [int, float],
        'FMAX': [int, float],
        'FPITCH': [int, float],
        'PDOS': [str],
        'PROJECTION_DIRECTION': [list],
        'XYZ_DIRECTION': [bool],
        'SIGMA': [int, float],
        'DEBYE_MODEL': [bool],
        'MOMEMT': [bool],
        'MOMENT_ORDER': [int],
        # Thermal properties and displacements related tags
        'TPROP': [bool],
        'TMIN': [int, float],
        'TMAX': [int, float],
        'TSTEP': [int, float],
        'PRETEND_REAL': [bool],
        'CUTOFF_FREQUENCY': [int, float],
        'TDISP': [bool],
        'TDISPMAT': [bool],
        'TDISPMAT_CIF': [int, float],
        # Specific q-points
        'QPOINTS': [list],
        'WRITEDM': [bool],
        # Non-analytical term correction
        'NAC_METHOD': [str],
        'Q_DIRECTION': [list],
        # Group velocity
        'GROUP_VELOCITY': [bool],
        'GV_DELTA_Q': [int, float],
        # Symmetry
        'SYMMETRY_TOLERANCE': [int, float],
        'SYMMETRY': [bool],
        'MESH_SYMMETRY': [bool],
        'FC_SYMMETRY': [bool],
        # Force constants
        'FULL_FORCE_CONSTANTS': [bool],
        'WRITE_FORCE_CONSTANTS': [bool],
        # Create animation file
        'ANIME_TYPE': [str],
        'ANIME': [list],
        # Create modulated structure
        'MODULATION': [str],
        # Characters of irreducible representations
        'IRREPS': [list],
        'SHOW_IRREPS': [bool],
        'LITTLE_COGROUP': [bool],
    }

    # Keywords that cannot be set by the user but will be set by the plugin
    _BLOCKED_TAGS = [
        'DIM',
        'ATOM_NAME',
        'MASS',  #??? from structure or also from here??
        'MAGMOM',
        'CREATE_DISPLACEMENTS',
        'DISPLACEMENT_DISTANCE',
        'DIAG',
        'PM',
        'RANDOM_DISPLACEMENTS',
        'RANDOM_SEED',
        'NAC',
        'FORCE_CONSTANTS',
        'READ_FORCE_CONSTANTS',
        'FC_FORMAT',
        'READFC_FORMAT',
        'WRITEFC_FORMAT',
        'BAND_FORMAT',
        'MESH_FORMAT',
        'QPOINTS_FORMAT',
        'HDF5',
    ]

    @classmethod
    def define(cls, spec):
        """Define inputs, outputs, and outline."""
        super().define(spec)

        # in v1.0 it will be a namespace of Dict
        spec.input(
            'parameters',
            valid_type=orm.Dict,
            required=True,
            help=(
                'Phonopy parameters (`setting tags`) for post processing. '
                'The following tags, along their type, are allowed:\n' +
                '\n'.join(f'{tag_name}' for tag_name in cls._AVAILABLE_TAGS)
            ),
            validator=cls._validate_parameters,
        )
        spec.input(
            'phonopy_data',
            valid_type=PhonopyData,
            required=False,
            help='The preprocess output info of a previous ForceConstantsWorkChain.'
        )
        spec.input(
            'force_constants',
            valid_type=ForceConstantsData,
            required=False,
            help='Force constants of the input structure.'
        )
        spec.input('settings', valid_type=orm.Dict, required=False, help='Settings for phonopy calculation.')
        spec.inputs.validator = cls._validate_inputs

        spec.input('metadata.options.withmpi', valid_type=bool, default=False)
        spec.input('metadata.options.input_filename', valid_type=str, default=cls._DEFAULT_INPUT_FILE)
        spec.input('metadata.options.output_filename', valid_type=str, default=cls._DEFAULT_OUTPUT_FILE)
        spec.input('metadata.options.parser_name', valid_type=str, default='phonopy.phonopy')
        spec.inputs['metadata']['options']['resources'].default = lambda: {'num_machines': 1}

        spec.output(
            'output_parameters', valid_type=orm.Dict, required=False, help='Sum up info of phonopy calculation.'
        )
        spec.output(cls._OUTPUTS['fc'][1], valid_type=orm.ArrayData, required=False, help='Calculated force constants.')
        spec.output(cls._OUTPUTS['dos'][1], valid_type=orm.XyData, required=False, help='Calculated total DOS.')
        spec.output(cls._OUTPUTS['pdos'][1], valid_type=orm.XyData, required=False, help='Calculated projected DOS.')
        spec.output(
            cls._OUTPUTS['band'][1], valid_type=orm.BandsData, required=False, help='Calculated phonon band structure.'
        )
        spec.output(cls._OUTPUTS['qpoints'][1], valid_type=orm.BandsData, required=False, help='Calculated qpoints.')
        spec.output(cls._OUTPUTS['mesh'][1], valid_type=orm.BandsData, required=False, help='Calculated qpoint mesh.')
        spec.output(
            cls._OUTPUTS['irreps'][1], valid_type=orm.Dict, required=False, help='Irreducible representation output.'
        )
        spec.output(
            cls._OUTPUTS['tprop'][1], valid_type=orm.XyData, required=False, help='Calculated thermal properties.'
        )
        spec.output(
            cls._OUTPUTS['tdisp'][1], valid_type=orm.Dict, required=False, help='Calculated thermal displacements.'
        )
        spec.output(
            cls._OUTPUTS['tdispmat'][1],
            valid_type=orm.Dict,
            required=False,
            help='Calculated thermal displacements matrices.'
        )
        spec.output(cls._OUTPUTS['mod'][1], valid_type=orm.Dict, required=False, help='Modulation information.')

        # Unrecoverable errors: required retrieved files could not be read, parsed or are otherwise incomplete
        spec.exit_code(
            301, 'ERROR_NO_RETRIEVED_TEMPORARY_FOLDER', message='The retrieved temporary folder could not be accessed.'
        )
        spec.exit_code(
            302,
            'ERROR_OUTPUT_STDOUT_MISSING',
            message='The retrieved folder did not contain the required stdout output file.'
        )
        spec.exit_code(
            303,
            'ERROR_OUTPUT_PHONOPY_MISSING',
            message='The retrieved folder did not contain the required phonopy file.'
        )
        spec.exit_code(
            303,
            'ERROR_OUTPUT_FILES_MISSING',
            message='The retrieved folder did not contain one or more expected output files.'
        )
        spec.exit_code(
            305, 'ERROR_BAD_INPUTS', message='No run mode has been selected.'
        )  # maybe we should check this from the input too

        spec.exit_code(310, 'ERROR_OUTPUT_STDOUT_READ', message='The stdout output file could not be read.')
        spec.exit_code(311, 'ERROR_OUTPUT_STDOUT_PARSE', message='The stdout output file could not be parsed.')
        spec.exit_code(
            312,
            'ERROR_OUTPUT_STDOUT_INCOMPLETE',
            message='The stdout output file was incomplete probably because the calculation got interrupted.'
        )

        spec.exit_code(320, 'ERROR_OUTPUT_YAML_LOAD', message='The loading of yaml file got an unexpected error.')
        spec.exit_code(321, 'ERROR_OUTPUT_NUMPY_LOAD', message='The file loading via numpy got an unexpected error.')

        spec.exit_code(350, 'ERROR_UNEXPECTED_PARSER_EXCEPTION', message='The parser raised an unexpected exception.')

        # Not implemented file parser
        spec.exit_code(
            400, 'ERROR_NOT_IMPLEMENTED_PARSER', message='The parser was not able to parse one or more files.'
        )

    @classmethod
    def _validate_parameters(cls, value, _):
        """Validate the ``parameters`` input namespace."""

        def __validate_dict(value_dict):
            enabled_dict = cls._AVAILABLE_TAGS
            unknown_tags = set(value_dict.keys()) - set(enabled_dict.keys())
            if unknown_tags:
                return (
                    f"Unknown tags in 'parameters': {unknown_tags}, "
                    f'allowed tags are {cls._AVAILABLE_TAGS.keys()}.'
                )
            invalid_values = [
                value_dict[key] for key in value_dict.keys() if not type(value_dict[key]) in enabled_dict[key]
            ]
            if invalid_values:
                return f'Parameters tags must be of the correct type; got invalid values {invalid_values}.'

        if value:
            if isinstance(value, orm.Dict):
                __validate_dict(value.get_dict())

    @classmethod
    def _validate_inputs(cls, value, _):
        """Validate the top level namespace."""
        if 'force_constants' in value or 'phonopy_data' in value:
            if 'force_constants' in value and 'phonopy_data' in value:
                return 'specify only one between `force_constants` and `phonopy_data`'
        else:
            return 'at least one between `force_constants` and `phonopy_data` must be specified'

    def prepare_for_submission(self, folder):
        """Prepare the calculation job for submission by transforming input nodes into input files.

        In addition to the input files being written to the sandbox folder, a `CalcInfo` instance will be returned that
        contains lists of files that need to be copied to the remote machine before job submission, as well as file
        lists that are to be retrieved after job completion.

        :param folder: a sandbox folder to temporarily write files on disk.
        :return: :py:`~aiida.common.datastructures.CalcInfo` instance.
        """
        retrieve_list = []
        retrieve_temporary_list = []
        calculation_cmds = []

        if 'settings' in self.inputs:
            settings = _lowercase_dict(self.inputs.settings.get_dict(), dict_name='settings')
        else:
            settings = {}
        # It can contain:
        # "keep_animation_files": bool
        # "symmetrize_nac": bool
        # "factor_nac": float
        # "subtract_residual_forces": bool

        # ================= prepare the phonopy input files ===================

        self.write_phonopy_info(folder)

        if 'force_constants' in self.inputs:
            self.write_force_constants(folder)

        # =============== prepare the submit and conf file ====================

        parameters = _uppercase_dict(self.inputs.parameters.get_dict(), dict_name='parameters')
        filename = self.inputs.metadata.options.input_filename
        self.write_calculation_input(folder, parameters, filename)

        # from phonopy recommendations, filename of the inputs configuration goes first
        calculation_cmds.append([filename])

        # ================= retreiving phonopy output files ===================

        # Keeping in the repository only the stdout, as deafult.
        # Animation files and `phonopy.yaml` can be retrieved preparing
        # the `settings` input namespace accordingly.
        retrieve_list.append(self.inputs.metadata.options.output_filename)
        if settings.pop('keep_animation_files', None):
            for format_ in ['jmol', 'xyz', 'xyz_jmol', 'arc', 'ascii']:
                retrieve_list.append(f'anime.{format_}')
            retrieve_list.append('APOSCAR-*')

        if settings.pop('keep_phonopy_yaml', False):
            retrieve_list.append(self._DEFAULT_PHONOPY_FILE)
        else:
            retrieve_temporary_list.append(self._DEFAULT_PHONOPY_FILE)

        # Retrieving everything and raising error from parser if something is missing.
        for value in self._OUTPUTS.values():
            retrieve_temporary_list.append(value[0])  # first value of the tuple
        if 'force_constants' in self.inputs:  # otherwise we retrieve the same thing as the input
            retrieve_temporary_list.pop(self._INPUT_FORCE_CONSTANTS)

        # ============================ calcinfo ===============================

        local_copy_list = []

        calcinfo = datastructures.CalcInfo()
        calcinfo.uuid = self.uuid
        calcinfo.local_copy_list = local_copy_list  # what we write in the folder
        calcinfo.retrieve_list = retrieve_list  # what to retrieve and keep
        # what to retrieve temporarly for just parsing purposes
        calcinfo.retrieve_temporary_list = retrieve_temporary_list

        calcinfo.codes_info = []
        for cmdline_params in calculation_cmds:
            codeinfo = datastructures.CodeInfo()  # new code info object per cmdline
            codeinfo.cmdline_params = cmdline_params
            codeinfo.stdout_name = self.inputs.metadata.options.output_filename
            codeinfo.code_uuid = self.inputs.code.uuid
            codeinfo.withmpi = self.inputs.metadata.options.withmpi
            # appending CodeInfo to CalcInfo
            calcinfo.codes_info.append(codeinfo)

        return calcinfo

    def _get_p2s_map(self):
        """Get the primitive to supercell map."""
        if 'force_constants' in self.inputs:
            return self.inputs.force_constants.get_cells_mappings()['primitive']['p2s_map']
        if 'phonopy_data' in self.inputs:
            return self.inputs.phonopy_data.get_cells_mappings()['primitive']['p2s_map']
        return None

    def write_phonopy_info(self, folder):
        """Write in `folder` the `phonopy.yaml` file."""
        from phonopy.interface.phonopy_yaml import PhonopyYaml

        kwargs = {}

        if 'settings' in self.inputs:
            the_settings = self.inputs.settings.get_dict()
            for key in ['symmetrize_nac', 'factor_nac', 'subtract_residual_forces']:
                if key in the_settings:
                    kwargs.update({key: the_settings[key]})

        if 'phonopy_data' in self.inputs:
            ph = self.inputs.phonopy_data.get_phonopy_instance(**kwargs)
        elif 'force_constants' in self.inputs:
            ph = self.force_constants.get_phonopy_instance(**kwargs)

        # Setting the phonopy yaml obtject to produce yaml lines
        # .. note: this does not write the force constants
        phpy_yaml = PhonopyYaml()
        phpy_yaml.set_phonon_info(ph)
        phpy_yaml_txt = str(phpy_yaml)

        with folder.open(self._DEFAULT_PHONOPY_FILE, 'w', encoding='utf8') as handle:
            handle.write(phpy_yaml_txt)

    def write_force_constants(self, folder):
        """Write in `folder` the force constants file."""
        from phonopy.file_IO import write_force_constants_to_hdf5

        filename = folder.get_abs_path(self._INPUT_FORCE_CONSTANTS)

        write_force_constants_to_hdf5(
            force_constants=self.inputs.force_constants.force_constants, filename=filename, p2s_map=self._get_p2s_map()
        )

    def write_calculation_input(self, folder, parameters: dict, filename: str):
        """Write in `folder` the input file containing the information regarding the calculation."""
        characters = []  # single characters to be written

        # appending eventual blocked keys to parameters
        if 'force_constants' in self.inputs:
            parameters.append({'FORCE_CONSTANTS': 'READ'})

        if 'nac_parameters' in self.inputs:
            parameters.append({'NAC': True})

        # if not "SYMMETRY_TOLERANCE" in parameters.keys():
        #     symprec = self.inputs.phonopy_data.symprec
        #     parameters.update({"SYMMETRY_TOLERANCE":symprec})

        # write potential huge outputs in `.hdf5` format, and read force constants from this format too
        parameters.update({'HDF5': True})

        # append single character
        for key, value in parameters.items():
            characters.append(key)
            characters.append('=')
            if isinstance(value, str):
                characters.append(value)
            elif isinstance(value, bool):
                characters.append(f'.{value}.')
            elif isinstance(value, (float, int)):
                characters.append(str(value))
            elif isinstance(value, list):
                string_value = ''
                for val in value:
                    string_value += str(val) + ' '
                characters.append(string_value[:-1])
            else:
                raise InputValidationError(f'value type `{type(value)}` is not supported.')
            characters.append('\n')  # one line for each key in parameters

        with folder.open(filename, 'w', encoding='utf8') as handle:
            for character in characters:
                handle.write(character)



# -*- coding: utf-8 -*-
"""Raw parsers of the phonopy output files."""

from aiida_phonopy.utils.mapping import get_logging_container


def parse_stdout(stdout):
    """Raw parser of the phonopy std output.

    :param stdout: str of the std output of phonopy
    :returns: tuple of two dictionaries, with the parsed data and log messages, respectively
    """
    stdout_lines = stdout.splitlines()
    logs = get_logging_container()

    parsed_data = {}
    is_incomplete = True

    for line in stdout_lines:
        if 'Python' in line:
            parsed_data.update({'python_version': line.split()[-1]})

        if 'Spglib' in line:
            parsed_data.update({'spglib_version': line.split()[-1]})

        if 'One of the following run modes may be specified for phonon calculations.' in line:
            logs.error.append('ERROR_BAD_INPUTS')

        if 'Summary of calculation was written in "phonopy.yaml".' in line:
            is_incomplete = False

    if is_incomplete:
        logs.error.append('ERROR_OUTPUT_STDOUT_INCOMPLETE')

    return parsed_data, logs



# -*- coding: utf-8 -*-
"""Parsers of `PhonopyCalculation` output files."""

import os
import traceback

from aiida import orm
import h5py
import numpy as np
import yaml

from aiida_phonopy.calculations.phonopy import PhonopyCalculation
from aiida_phonopy.utils.mapping import _uppercase_dict, get_logging_container

from .base import Parser


def file_opener(folder_path, filename):
    """Function to open sistematically each expected output format."""
    filename_path = os.path.join(folder_path, filename)

    if filename.endswith('hdf5'):
        return h5py.File(filename_path, 'r')

    return open(filename_path, mode='r', encoding='utf-8')


class PhonopyParser(Parser):
    """Parser the files produced by a phonopy post processing calculation."""

    def parse(self, **kwargs):
        """Parse retrieved files from remote folder."""
        retrieved = self.retrieved
        retrieve_temporary_list = self.node.base.attributes.get('retrieve_temporary_list', None)

        # If temporary files were specified, check that we have them
        if retrieve_temporary_list:
            try:
                retrieved_temporary_folder = kwargs['retrieved_temporary_folder']
            except KeyError:
                return self.exit(self.exit_codes.ERROR_NO_RETRIEVED_TEMPORARY_FOLDER)

            # !FIXME!
            # We should implement `filenames` depending whether they were
            # put in the retrieved or in the temporary folder.
            filenames = os.listdir(retrieved_temporary_folder)

        # Parse the stdout
        parsed_stdout, logs_stdout, exit_code_stdout = self.parse_stdout()
        self.emit_logs(logs_stdout)

        if exit_code_stdout:
            return self.exit(exit_code_stdout)

        # Parse phonopy.yaml
        try:
            filename = PhonopyCalculation._DEFAULT_PHONOPY_FILE
            filenames.remove(filename)

            with file_opener(retrieved_temporary_folder, filename) as file:
                try:
                    parsed_phonopy_yaml = yaml.safe_load(file)
                except yaml.YAMLError:
                    return self.exit(self.exit_codes.ERROR_OUTPUT_YAML_LOAD)
        except ValueError:
            try:
                with retrieved.base.repository.open(filename) as file:
                    try:
                        parsed_phonopy_yaml = yaml.safe_load(file)
                    except yaml.YAMLError:
                        return self.exit(self.exit_codes.ERROR_OUTPUT_YAML_LOAD)
            except IOError:
                # This should in principle never happen!
                return self.exit(self.exit_codes.ERROR_OUTPUT_PHONOPY_MISSING)

        # Output parameters
        phonopy_yaml_keys = ['phonopy', 'physical_unit', 'space_group']
        parsed_parameters = {key: value for key, value in parsed_phonopy_yaml.items() if key in phonopy_yaml_keys}
        parsed_parameters.update(parsed_stdout)
        self.out('output_parameters', orm.Dict(parsed_parameters))

        # We use the input `parameters` to check the expected retrieved files in folder.
        expected_filenames_keys = self.get_expected_filenames_keys()

        parsers = {
            'fc': self.parse_force_constants,
            'dos': self.parse_total_dos,
            'pdos': self.parse_projected_dos,
            'band': self.parse_band_structure,
            'mesh': self.parse_qpoints,  # to do a proper one
            'qpoints': self.parse_qpoints,  # to do a proper one
            'irreps': self.parse_yaml,  # to do a proper one
            'tprop': self.parse_thermal_properties,
            'tdisp': self.parse_yaml,  # to do a proper one
            'tdispmat': self.parse_yaml,  # to do a proper one
        }

        missing_files = []

        # Loop over every expected output file, parse it and expose its outputs.
        # If one or more are missing, we put it in the `missing_files` and raise error afterwards.
        for expected_key in expected_filenames_keys:
            try:
                filename, filout = PhonopyCalculation._OUTPUTS[expected_key]
                filenames.remove(filename)
                if expected_key == 'fc':
                    fc_output = self.parse_force_constants(os.path.join(retrieved_temporary_folder, filename))
                    self.out(filout, fc_output)
                else:
                    with file_opener(retrieved_temporary_folder, filename) as file:
                        self.out(filout, parsers[expected_key](file))
            except (FileNotFoundError, ValueError):
                missing_files.append(filename)

        if missing_files:
            return self.exit(self.exit_codes.ERROR_OUTPUT_FILES_MISSING)

        # We don't parse animation files.
        for filename in filenames:
            if filename.endswith(('jmol', 'xyz', 'xyz_jmol', 'arc', 'ascii')) or filename.startswith('APOSCAR-*'):
                filenames.remove(filename)

        # !FIXME!
        # Still to work out better. Good chances are that is not necessary...
        #
        # If something is still present and it has not been parsed, we raise error.
        # if filenames:
        #     self.exit_codes.ERROR_NOT_IMPLEMENTED_PARSER

    def parse_stdout(self):
        """Parse the stdout output file.

        :param parameters: the input parameters dictionary
        :param parsed_phonopy: the collected parsed data from the yaml phonopy output
        :return: raw parsed data
        """
        from .raw_parsers.phonopy import parse_stdout as parse_stdout_

        parsed_data = {}
        exit_code_stdout = None

        filename_stdout = self.node.get_option('output_filename')
        logs = get_logging_container()  # datastructure for logs

        if filename_stdout not in self.retrieved.base.repository.list_object_names():
            exit_code_stdout = self.exit_codes.ERROR_OUTPUT_STDOUT_MISSING
            return parsed_data, logs, exit_code_stdout

        try:
            stdout = self.retrieved.base.repository.get_object_content(filename_stdout)
        except (IOError, OSError):
            exit_code_stdout = self.exit_codes.ERROR_OUTPUT_STDOUT_READ
            return parsed_data, logs, exit_code_stdout

        try:
            # Check for job completion, indicating that phonopy exited without interruption, even if there was an error.
            parsed_data, logs = parse_stdout_(stdout)
        except ValueError:
            logs.critical.append(traceback.format_exc())
            exit_code_stdout = self.exit_codes.ERROR_UNEXPECTED_PARSER_EXCEPTION

        # If the stdout was incomplete, most likely the job was interrupted before it could cleanly finish, so the
        # output files are most likely corrupt and cannot be restarted from
        if 'ERROR_OUTPUT_STDOUT_INCOMPLETE' in logs['error']:
            exit_code_stdout = self.exit_codes.ERROR_OUTPUT_STDOUT_INCOMPLETE
        if 'ERROR_BAD_INPUTS' in logs['error']:
            exit_code_stdout = self.exit_codes.ERROR_BAD_INPUTS

        return parsed_data, logs, exit_code_stdout

    def get_expected_filenames_keys(self):
        """Return the retrieve file keys (that map to the filenames outputs) depending on the tags in `parameters`."""
        retrieved_set = set()
        parameters = _uppercase_dict(self.node.inputs.parameters.get_dict(), dict_name='parameters')

        maps = {
            'band': {'BAND', 'BAND_PATHS', 'BAND_POINTS', 'BAND_LABELS', 'BAND_CONNECTION', 'BAND_INDICES'},
            'mesh': {'MESH', 'MP', 'MESH_NUMBERS', 'MP_SHIFT', 'GAMMA_CENTER', 'WRITE_MESH'},
            'dos': {'DOS'},
            'pdos': {'PDOS'},
            'tprop': {'TPROP'},
            'tdisp': {'TDISP'},
            'tdispmat': {'TDISPMAT'},
            'qpoints': {'QPOINTS', 'WRITEDM'},
            'fc': {'WRITE_FORCE_CONSTANTS'},
            'mod': {'MODULATION'},
            'irreps': {'IRREPS', 'SHOW_IRREPS', 'LITTLE_COGROUP'}
        }

        for tag, value in parameters.items():
            for key_map, tag_map in maps.items():
                if tag in tag_map:
                    # Usually these tags are set only if one needs them, i.e. to TRUE,
                    # thus this further control is actually pointless. Nevertheless,
                    # we want to avoid anyway to raise error when in inputs one of these
                    # tags is set to FALSE, meaning there will be no output to parse.
                    if tag.startswith(('WRITE', 'DOS', 'TPROP', 'TDISP', 'TDISPMAT')):
                        if value:
                            retrieved_set.add(key_map)
                    else:
                        retrieved_set.add(key_map)

        # Double check on mesh. The writing tag must have priority.
        if 'WRITE_MESH' in parameters.keys():
            if not parameters['WRITE_MESH']:
                retrieved_set.remove('mesh')

        return retrieved_set

    def parse_force_constants(self, filepath):
        """Parse the `force_constants.hdf5` output file."""
        from phonopy.file_IO import read_force_constants_hdf5

        p2s_map = self._get_p2s_map()
        force_constants = read_force_constants_hdf5(filename=filepath, p2s_map=p2s_map)

        fc_array = orm.ArrayData()
        fc_array.set_array('force_constants', force_constants)
        fc_array.label = 'force_constants'

        return fc_array

    def load_with_numpy(self, file):
        """Load a txt file using numpy."""
        try:
            data = np.loadtxt(file)
        except ValueError:
            self.exit(self.exit_codes.ERROR_OUTPUT_NUMPY_LOAD)
        return data

    def load_with_yaml(self, file):
        """Load a yaml file using."""
        try:
            data = yaml.safe_load(file)
        except yaml.YAMLError:
            self.exit(self.exit_codes.ERROR_OUTPUT_YAML_LOAD)
        return data

    def parse_yaml(self, file):
        """Parse a `.yaml` file and return it as a Dict."""
        data = self.load_with_yaml(file=file)
        return orm.Dict(data)

    def parse_total_dos(self, file):
        """Parse `total_dos.dat` output file."""
        data = self.load_with_numpy(file=file)

        total_dos = {'frequency_points': data[:, 0], 'total_dos': data[:, 1]}
        dos = orm.XyData()
        dos.set_x(total_dos['frequency_points'], 'Frequency', 'THz')
        dos.set_y(total_dos['total_dos'], 'DOS', '1/THz')
        dos.label = 'Total DOS'

        return dos

    def parse_projected_dos(self, file):
        """Parse `projected_dos.dat` output file."""
        data = self.load_with_numpy(file=file)

        projected_dos = {'frequency_points': data[:, 0], 'projected_dos': data[:, 1:].T}
        pdos = orm.XyData()
        pdos_list = list(projected_dos['projected_dos'])
        pdos.set_x(projected_dos['frequency_points'], 'Frequency', 'THz')
        pdos.set_y(
            pdos_list,
            [
                'Projected DOS',
            ] * len(pdos_list),
            [
                '1/THz',
            ] * len(pdos_list),
        )
        pdos.label = 'Projected DOS'

        return pdos

    def parse_thermal_properties(self, file):
        """Parse the `thermal_properties.yaml`` output file."""
        data = self.load_with_yaml(file=file)

        thermal_properties = {
            'temperatures': [],
            'free_energy': [],
            'entropy': [],
            'heat_capacity': [],
        }

        for tp in data['thermal_properties']:
            thermal_properties['temperatures'].append(tp['temperature'])
            thermal_properties['entropy'].append(tp['entropy'])
            thermal_properties['free_energy'].append(tp['free_energy'])
            thermal_properties['heat_capacity'].append(tp['heat_capacity'])
        old_thermal_property = thermal_properties.copy()
        for key, item in old_thermal_property.items():
            thermal_properties[key] = np.array(item)

        tprops = orm.XyData()
        tprops.set_x(thermal_properties['temperatures'], 'Temperature', 'K')
        tprops.set_y(
            [
                thermal_properties['free_energy'],
                thermal_properties['entropy'],
                thermal_properties['heat_capacity'],
            ],
            ['Helmholtz free energy', 'Entropy', 'Cv'],
            ['kJ/mol', 'J/K/mol', 'J/K/mol'],
        )
        tprops.label = 'Thermal properties'

        return tprops

    def parse_band_structure(self, file, freqs_units='THz'):
        """Parse the `band.hdf5`` output file.

        Expected keys are:
            * **nqpoint**: array with total number of q-points in the band structure, array(1,)
            * **frequency**: array of frequencies at each q-point; array(npath, segment_nqpoint, nband)
            * **label**: array of labels, two per each path; array(npath, 2)
            * **path**: number of q-points per path; array(npath, segment_nqpoint, qpoint/qposition(3,))
            * **distance**: distance between consecutive q-points in a segment of path; array(npath, segment_nqpoint)
            * **segment_nqpoint**: number of q-points per segment of the entire path; array(npath,)
            * **eigenvector**: (optional) eigenvector of each phonon mode, i.e. each frequency of each q-point;
                array(npath, segment_nqpoint, nband, eigenvector(6,))
            * **group_velocity**: (optional) group velocity of each phonon mode (as for eigenvector);
                array(npath, segment_nqpoint, nband, group_velocity(3,))
        """
        import re

        # Loading the data
        data = {key: file[key][:] for key in file.keys()}

        # Initializing the bands object and setting the reciprocal lattice from the unitcell.
        if 'phonopy_data' in self.node.inputs:
            structure = self.node.inputs['phonopy_data'].get_unitcell()
        else:
            structure = self.node.inputs['force_constants'].get_unitcell()

        band_structure = orm.BandsData()
        band_structure.set_cell_from_structure(structuredata=structure)

        # Extracting the frequencies and qpoints.
        frequencies = []
        qpoints = []
        for freqs_per_path, qpoint_per_path in zip(data['frequency'], data['path']):
            for freqs_per_qpoint in freqs_per_path:
                frequencies.append(freqs_per_qpoint)
            for qpoint in qpoint_per_path:
                qpoints.append(qpoint)

        # Extracting the labels along with the index of the qpoint.
        labels = []
        nqpoint = [0, -1]
        for label_pair, segment in zip(
            data['label'], data['segment_nqpoint']
        ):  # they are saved in pairs, two per segment
            nqpoint[1] += segment
            for i, label in enumerate(label_pair):
                try:  # the labels are saved in dtype |S16 --> e.g. b'$\\mathrm{X}$'
                    symbol = re.search(r'\{(\w+)\}', str(label)).group(1)
                except AttributeError:  # it means we hit Gamma
                    if label == b'$\\Gamma$':
                        symbol = r'$\Gamma$'
                    else:
                        symbol = '?'
                labels.append([nqpoint[i], symbol])
            nqpoint[0] += segment

        # Here I remove the symilar symbols if they are consecutive.
        # The sorting could be smarter, but it works.
        final_labels = []
        for nq, pair in enumerate(labels):
            if not nq == 0:
                if not pair[1] == labels[nq - 1][1]:
                    final_labels.append(pair)
            else:
                final_labels.append(pair)

        # Setting the final bands structure data.
        band_structure.set_kpoints(np.array(qpoints))
        band_structure.set_bands(np.array(frequencies), units=freqs_units)
        band_structure.labels = final_labels

        return band_structure

    def parse_qpoints(self, file, freqs_units='THz'):
        """Parse the `mesh.hdf5`` and `qpoints.hdf5`` output files.

        Expected keys are:
            * **frequency**: array of frequencies at each q-point; array(nqpoint, nband)
            * **mesh**: qpoint mesh; array(3,)
            * **qpoint**: qpoints; array(nqpoint, 3)
            * **weight**: weight of qpoints; array(nqpoint,)
            * **eigenvector**: (optional) eigenvector of each phonon mode, i.e. each frequency of each q-point;
                array(npath, segment_nqpoint, nband, eigenvector(6,))
            * **group_velocity**: (optional) group velocity of each phonon mode (as for eigenvector);
                array(npath, segment_nqpoint, nband, group_velocity(3,))
        """
        # Loading the data
        data = {key: file[key][:] for key in file.keys()}

        # Initializing the bands object and setting the reciprocal lattice from the unitcell.
        if 'phonopy_data' in self.node.inputs:
            structure = self.node.inputs['phonopy_data'].get_unitcell()
        else:
            structure = self.node.inputs['force_constants'].get_unitcell()

        band_structure = orm.BandsData()
        band_structure.set_cell_from_structure(structuredata=structure)

        qpoints = np.array(data['qpoint'])
        frequencies = np.array(data['frequency'])
        if file.filename.endswith('qpoints.hdf5'):
            weights = None
        else:
            weights = np.array(data['weight'].astype(float))

        # Setting the final bands structure data.
        band_structure.set_kpoints(qpoints, weights=weights)
        band_structure.set_bands(frequencies, units=freqs_units)

        return band_structure

    def _get_p2s_map(self):
        """Get the primitive to supercell map."""
        if 'force_constants' in self.node.inputs:
            return self.node.inputs.force_constants.get_cells_mappings()['primitive']['p2s_map']
        if 'phonopy_data' in self.node.inputs:
            return self.node.inputs.phonopy_data.get_cells_mappings()['primitive']['p2s_map']
        return None



# -*- coding: utf-8 -*-
"""
This module defines the base class for other Data types.
"""

from copy import deepcopy

from aiida import orm
from aiida.orm.nodes.data import ArrayData
import numpy as np
from phonopy import Phonopy
from phonopy.structure.cells import PhonopyAtoms

from aiida_phonopy.calculations.functions.link_structures import (
    phonopy_atoms_from_structure,
    phonopy_atoms_to_structure,
)


def _get_valid_matrix(matrix):
    """Get and validate the `supercell_matrix` and `primitive_matrix` inputs.

    :param matrix: (3,1) or (3,3) shape array
    :type matrix: list, orm.List, numpy.ndarray
    :return: a (3,3) numpy.ndarray
    :raises:
        * TypeError: if it is not a valid array type and if the array does not contain only numbers
        * ValueError: if the array is not of the correct shape
    """
    if not isinstance(matrix, (list, orm.List, np.ndarray)):
        raise TypeError("only `list`, `aiida.orm.List` and `numpy.ndarray` (and 'auto' as string for primitive)")

    if isinstance(matrix, np.ndarray):
        matrix = matrix.tolist()

    if not len(matrix) == 3:
        raise ValueError('need exactly (3,1) or (3,3) shape.')

    for row in matrix:
        if isinstance(row, list):
            if not len(row) in [0, 3]:
                raise ValueError('matrix need to have (3,1) or (3,3) shape.')
            for element in row:
                if not isinstance(element, (int, float)):
                    raise TypeError(
                        f'type `{type(element)}` of {element} is not an accepted '
                        'type in matrix; only `int` and `float` are valid.'
                    )

    valid_matrix = np.array(matrix)
    valid_matrix = np.diag(valid_matrix) if valid_matrix.shape == (3,) else valid_matrix

    return valid_matrix


class RawData(ArrayData):  # pylint: disable=too-many-ancestors
    """
    This class contaings the base information for the other phonon and phonopy related data types,
    i.e. the unitcell and its symmetry information, the supercell and primitive cell matrices.
    """

    def __init__(
        self,
        structure: orm.StructureData = None,
        phonopy_atoms: PhonopyAtoms = None,
        supercell_matrix: list = None,
        primitive_matrix: list = None,
        symprec: float = 1e-05,
        is_symmetry: bool = True,
        distinguish_kinds: bool = True,
        **kwargs
    ):
        if structure and phonopy_atoms:
            raise ValueError('cannot pass `structure`and `phonopy_atoms` at the same time')

        if structure:
            phonopy_atoms_unitcell, mapping = phonopy_atoms_from_structure(structure)
        elif phonopy_atoms:
            phonopy_atoms_unitcell, mapping = (phonopy_atoms, None)
        else:
            raise ValueError('at least one between `structure` and `phonopy_atoms` must be specified')

        args = {
            'unitcell': phonopy_atoms_unitcell,
            'supercell_matrix': supercell_matrix,
            'primitive_matrix': primitive_matrix,
            'symprec': symprec,
            'is_symmetry': is_symmetry,
        }

        try:
            Phonopy(**args)
        except Exception as err:
            raise ValueError('one or more inputs are not consistent with the `phonopy.Phonopy` format') from err

        super().__init__(**kwargs)

        # Inizializing the class attributes.
        self._set_phonopy_version()
        self._set_unitcell_attributes(phonopy_atoms=phonopy_atoms_unitcell)
        self._set_kinds_map(mapping)
        self._set_distinguish_kinds(distinguish_kinds)  # crucial before setting symbols/names!
        self._set_symbols_and_names()

        if not supercell_matrix is None:
            self._set_supercell_matrix(supercell_matrix)
        else:
            self._set_supercell_matrix(np.eye(3))

        self._set_symprec(symprec)
        self._set_is_symmetry(is_symmetry)

        # Important that is the last to be specified since, if "auto" is passed,
        # a phonopy instance is generated to get the primitive structure.
        if not primitive_matrix is None:
            self._set_primitive_matrix(primitive_matrix)
        else:
            self._set_primitive_matrix('auto')

    @property
    def phonopy_version(self):
        """Get the Phonopy version used."""
        return self.base.attributes.get('phonopy_version')

    def _set_phonopy_version(self):
        """Set the installed Phonopy version."""
        from phonopy.version import __version__ as the_phonopy_version

        self._if_can_modify()
        self.base.attributes.set('phonopy_version', the_phonopy_version)

    @property
    def numbers(self):
        """Get the `numbers` array of the atoms in the cell."""
        return self.base.attributes.get('numbers')

    @property
    def masses(self):
        """Get the `masses` array of the atoms in the cell."""
        return self.base.attributes.get('masses')

    @property
    def positions(self):
        """Get the `positions` array of the atoms in the cell."""
        return self.base.attributes.get('positions')

    @property
    def cell(self):
        """Get the `cell` of the structure."""
        return self.base.attributes.get('cell')

    @property
    def magnetic_moments(self):
        """Get the `magnetic_moments` array of the atoms in the cell."""
        return self.base.attributes.get('magnetic_moments')

    @property
    def symbols(self):
        """Get the chemical `symbols` array of the atoms in the cell."""
        return self.base.attributes.get('symbols')

    @property
    def pbc(self):
        """Get the periodic boundary conditions."""
        return self.base.attributes.get('pbc')

    @property
    def names(self):
        """Get the custom `names` array of the atoms in the cell.

        .. note: if no special names are specified, this will be equal to `symbols`."""
        return self.base.attributes.get('names')

    def _set_unitcell_attributes(self, phonopy_atoms: PhonopyAtoms):
        """Set the attributes for full reproducibility of the `PhonopyAtoms` class."""
        self._if_can_modify()
        self.base.attributes.set_many({
            'numbers': phonopy_atoms.numbers,
            'masses': phonopy_atoms.masses,
            'positions': phonopy_atoms.positions,
            'cell': phonopy_atoms.cell,
            'magnetic_moments': phonopy_atoms.magnetic_moments,
            'pbc': [True, True, True],  # to change when modifying Phonopy
        })

    def _set_symbols_and_names(self):
        """Set the `symbols` and `names`."""
        if self.kinds_map:
            numbers_to_names, numbers_to_symbols = self.kinds_map
            symbols = []
            names = []
            for number in self.numbers:
                try:
                    names.append(numbers_to_names[number])
                    symbols.append(numbers_to_symbols[number])
                except KeyError:
                    names.append(numbers_to_names[str(number)])
                    symbols.append(numbers_to_symbols[str(number)])
            self.base.attributes.set('symbols', symbols)
            self.base.attributes.set('names', names)
        else:
            phonopy_atoms = self._get_phonopy_atoms_unitcell(distinguish_kinds=True)
            self.base.attributes.set('symbols', deepcopy(phonopy_atoms.symbols))
            self.base.attributes.set('names', deepcopy(phonopy_atoms.symbols))

    @property
    def supercell_matrix(self):
        """Get the `supercell_matrix`."""
        return self.base.attributes.get('supercell_matrix')

    def _set_supercell_matrix(self, value):
        """Set the Phonopy supercell matrix.

        :param value: (3,3) or (3,1) shape array
        :type value: list, orm.List, numpy.ndarray
        :raises ModificationNotAllowed: if object is already stored
        """
        self._if_can_modify()
        the_supercell_matrix = _get_valid_matrix(value)

        self.base.attributes.set('supercell_matrix', deepcopy(the_supercell_matrix))

    @property
    def primitive_matrix(self):
        """Get the `primitive_matrix`."""
        return self.base.attributes.get('primitive_matrix')

    def _set_primitive_matrix(self, value):
        """Set the Phonopy primitive matrix.

        :param value: (3,3) or (3,1) shape array, or str
        :type value: list, orm.List, numpy.ndarray, str
        :raises ModificationNotAllowed: if object is already stored
        """
        self._if_can_modify()

        if isinstance(value, str):
            the_primitive_matrix = self.get_phonopy_instance(**{'primitive': value}).primitive_matrix
        else:
            the_primitive_matrix = _get_valid_matrix(value)

        self.base.attributes.set('primitive_matrix', deepcopy(the_primitive_matrix))

    @property
    def symprec(self):
        """Get the Phonopy symmetry tolerance (`symprec`)."""
        return self.base.attributes.get('symprec')

    def _set_symprec(self, value):
        """Set the Phonopy symmetry tolerance (`symprec`).

        :param value: tolerance for symmetry analysis. Check that you get
        the right symmetry of your structure before starting any calculation. Default is 1e-05.
        :type value: float
        :raises:
            * ModificationNotAllowed: if object is already stored
            * TypeError: if the input is not of type `float`
        """
        self._if_can_modify()
        if not isinstance(value, float):
            raise ValueError('only `float` type is accepted.')

        self.base.attributes.set('symprec', deepcopy(value))

    @property
    def is_symmetry(self):
        """Get `is_symmetry` value."""
        return self.base.attributes.get('is_symmetry')

    def _set_is_symmetry(self, value):
        """Set whether to use the symmetries. Use with care and if you know what your are doing.

        :param value: whether to use or not the symmetries. Deafault is True.
        :type value: bool
        :raises:
            * ModificationNotAllowed: if object is already stored
            * TypeError: if the input is not of type `bool`
        """
        self._if_can_modify()
        if not isinstance(value, bool):
            raise ValueError('only `bool` type is accepted.')

        self.base.attributes.set('is_symmetry', deepcopy(value))

    @property
    def kinds_map(self):
        """Get the map bewtween of the `numbers` and the `symbols` and `names`."""
        try:
            the_map = self.base.attributes.get('kinds_map')
        except AttributeError:
            the_map = None
        return the_map

    def _set_kinds_map(self, value):
        """Set the kind names map between the PhonopyAtoms unitcell and a reference
        structure. This is needed since PhonopyAtoms does not support kind names.
        This attribute allows to get proper `StructureData` supercells with displacements.

        :param value: tuple with two dictionaries (numbers_to_names, numbers_to_symbols)
        """
        self._if_can_modify()
        self.base.attributes.set('kinds_map', value)

    @property
    def distinguish_kinds(self):
        """Get whether or not kinds with same chemical symbol will be distinguished by symmetry."""
        return self.base.attributes.get('distinguish_kinds')

    def _set_distinguish_kinds(self, value):
        """Set whether or not to distinguish kinds."""
        self._if_can_modify()
        if not isinstance(value, bool):
            raise ValueError('only `bool` type is accepted.')

        self.base.attributes.set('distinguish_kinds', value)

    def _get_phonopy_atoms_unitcell(self, distinguish_kinds: bool) -> PhonopyAtoms:
        """Get the PhonopyAtoms object using the internal attributes."""
        kwargs = {
            'cell': self.cell,
            'positions': self.positions,
            'masses': self.masses,
            'magmoms': self.magnetic_moments
        }
        if distinguish_kinds:
            kwargs.update({'numbers': self.numbers})
        else:
            kwargs.update({'symbols': self.symbols})

        return PhonopyAtoms(**kwargs)

    def get_phonopy_instance(self, symmetrize_nac=None, factor_nac=None, **kwargs) -> Phonopy:
        """Return a `phonopy.Phonopy` object with the current values.

        :param symmetrize_nac: whether or not to symmetrize the nac parameters using point group symmetry.
        :type symmetrize_nac: bool, defaults to self.is_symmetry
        :param factor_nac: factor for non-analytical corrections
        :type factor_nac: float,defaults to Hartree*Bohr
        :param kwargs: for internal use to set the primitive cell
        """
        from phonopy.structure.symmetry import symmetrize_borns_and_epsilon
        from phonopy.units import Bohr, Hartree

        distinguish = kwargs.pop('distinguish', self.distinguish_kinds)
        unitcell = self._get_phonopy_atoms_unitcell(distinguish)

        primitive_matrix = kwargs.pop('primitive', None)
        if primitive_matrix is None:
            primitive_matrix = self.primitive_matrix

        ph_instance = Phonopy(
            unitcell=unitcell,
            supercell_matrix=self.supercell_matrix,
            primitive_matrix=primitive_matrix,
            symprec=self.symprec,
            is_symmetry=self.is_symmetry
        )

        # Non-analytical parameters
        if not isinstance(symmetrize_nac, bool) and symmetrize_nac is not None:
            raise TypeError('`symmetrize_nac` not of the right type')

        if symmetrize_nac is not None:
            do_symmetrize = symmetrize_nac
        else:
            do_symmetrize = self.is_symmetry

        if not isinstance(factor_nac, (float, int)) and factor_nac is not None:
            raise TypeError('input not of the right type')

        if factor_nac is not None:
            the_factor = factor_nac
        else:
            the_factor = Hartree * Bohr

        if self.has_nac_parameters():
            if do_symmetrize:
                nac = symmetrize_borns_and_epsilon(
                    borns=self.born_charges, epsilon=self.dielectric, ucell=ph_instance.primitive, symprec=self.symprec
                )
                the_born = nac[0]
                the_dielectric = nac[1]
            else:
                the_born = self.born_charges
                the_dielectric = self.dielectric

            ph_instance.nac_params = {'born': the_born, 'dielectric': the_dielectric, 'factor': the_factor}

        return ph_instance

    def get_unitcell(self) -> orm.StructureData:
        """Get the `unitcell` as StructureData (not stored)."""
        kwargs = {'distinguish': True}
        phonopy_instance = self.get_phonopy_instance(**kwargs).unitcell
        return phonopy_atoms_to_structure(phonopy_instance, self.kinds_map)

    def get_primitive_cell(self) -> orm.StructureData:
        """Get the `primitive cell` as StructureData (not stored)."""
        kwargs = {'distinguish': True}
        phonopy_instance = self.get_phonopy_instance(**kwargs).primitive
        return phonopy_atoms_to_structure(phonopy_instance, self.kinds_map)

    def get_supercell(self) -> orm.StructureData:
        """Get the pristine `supercell` as StructureData (not stored)."""
        kwargs = {'distinguish': True}
        phonopy_instance = self.get_phonopy_instance(**kwargs).supercell
        return phonopy_atoms_to_structure(phonopy_instance, self.kinds_map)

    def get_cells_mappings(self):
        """Return a dictionary containing the mappings among unit-, super- and primitive cell"""
        ph = self.get_phonopy_instance()

        cells_maps = {
            'primitive': {
                'p2p_map': ph.primitive.p2p_map,
                'p2s_map': ph.primitive.p2s_map,
                's2p_map': ph.primitive.s2p_map,
            },
            'supercell': {
                'u2u_map': ph.supercell.u2u_map,
                'u2s_map': ph.supercell.u2s_map,
                's2u_map': ph.supercell.s2u_map,
            },
        }

        return cells_maps

    @property
    def dielectric(self):
        """Get the `infinity` dielectric tensor (i.e. in the static field limit) in Cartesian coordinates."""
        try:
            value = self.base.attributes.get('dielectric')
            value = np.array(value)
        except (KeyError, AttributeError):
            value = None
        return value

    def set_dielectric(self, dielectric):
        """Set the `infinity` dielectric tensor in Cartesian coordinates.

        .. note: it is assumed that the reference system is the same (if not the) of the primitive cell.

        :param dielectric: (3, 3) array like

        :raises:
            * TypeError: if the format is not compatible or of the correct type
            * ValueError: if the format is not compatible or of the correct type
        """
        self._if_can_modify()

        if not isinstance(dielectric, (list, np.ndarray)):
            raise TypeError('the input is not of the correct type')

        the_dielectric = np.array(dielectric)

        if the_dielectric.shape == (3, 3):
            self.base.attributes.set('dielectric', the_dielectric.tolist())
        else:
            raise ValueError('the array is not of the correct shape')

    @property
    def born_charges(self):
        """Get the effective Born charges tensors in Cartesian coordinates."""
        try:
            value = self.get_array('born_charges')
        except (KeyError, AttributeError):
            value = None
        return value

    def set_born_charges(self, born_charges):
        """Set the `infinity` dielectric tensor in Cartesian coordinates.

        .. note: it is assumed that the tensors have been computed on the primitive cell.

        :param dielectric: (number of atoms in the primitive cell, 3, 3) shape array like

        :raises:
            * TypeError: if the format is not compatible or of the correct type
            * ValueError: if the format is not compatible or of the correct type
        """
        self._if_can_modify()

        if not isinstance(born_charges, (list, np.ndarray)):
            raise TypeError('the input is not of the correct type')

        the_born_charges = np.array(born_charges)

        nprimitive_atoms = len(self.get_primitive_cell().sites)

        if the_born_charges.shape == (nprimitive_atoms, 3, 3):
            self.set_array('born_charges', the_born_charges)
        else:
            raise ValueError('the array is not of the correct shape')

    def has_nac_parameters(self):
        """Returns wheter or not the Data has non-analytical constants."""
        return (self.dielectric is not None and self.born_charges is not None)

    def _if_can_modify(self):
        """Check if the object is stored and raise an error if so. To use in every setter."""
        from aiida.common.exceptions import ModificationNotAllowed

        if self.is_stored:
            raise ModificationNotAllowed('The PhonopyData object cannot be modified, it has already been stored')



# -*- coding: utf-8 -*-
"""
This module defines the class for managing the frozen phonon structure.
"""

import copy

from aiida import orm
import numpy as np

from aiida_phonopy.calculations.functions.link_structures import phonopy_atoms_to_structure

from .raw import RawData


class PreProcessData(RawData):  # pylint: disable=too-many-ancestors
    """
    This class is designed for handling the pre-process information regarding
    a frozen phonon calculation using `Phonopy`. These regard the unitcell structure,
    the supercell and primitive matrix, as well as other symmetry information.
    """

    def __init__(
        self,
        structure=None,
        phonopy_atoms=None,
        supercell_matrix=None,
        primitive_matrix=None,
        symprec=1e-05,
        is_symmetry=True,
        distinguish_kinds=True,
        **kwargs
    ):
        kwargs['structure'] = structure
        kwargs['phonopy_atoms'] = phonopy_atoms
        kwargs['supercell_matrix'] = supercell_matrix
        kwargs['primitive_matrix'] = primitive_matrix
        kwargs['symprec'] = symprec
        kwargs['is_symmetry'] = is_symmetry
        kwargs['distinguish_kinds'] = distinguish_kinds

        super().__init__(**kwargs)

        if self.__class__.__name__ == 'PreProcessData':
            self.set_displacements()

    @property
    def displacement_dataset(self):
        """Get the dispacement dataset in a readible format for phonopy."""
        try:
            the_displacement_dataset = self.base.attributes.get('displacement_dataset')
        except AttributeError:
            the_displacement_dataset = None
        return the_displacement_dataset

    @property
    def displacements(self):
        """Get the displacements to apply to the supercell.

        :returns: array with displacements; can be type-I or type-II (see :func:`phonopy.Phonopy.displacements`)
        """
        try:
            ph = self.get_phonopy_instance()
            ph.dataset = self.displacement_dataset

            return ph.displacements
        except (AttributeError, TypeError):
            return None

    def get_displacements(self):
        """Get the displacements to apply to the supercell."""
        return self.displacements

    def set_displacements(
        self,
        distance=0.01,
        is_plusminus='auto',
        is_diagonal=True,
        is_trigonal=False,
        number_of_snapshots=None,
        random_seed=None,
        temperature=None,
        cutoff_frequency=None,
    ):
        """Set displacements for frozen phonon calculation.
        (see `phonopy.Phonopy.generate_displacements` for a complete description of the inputs)

        .. note: temperature different from 0K can be given only when forces are set up.
        Thus, in the PreProcessData value different from zero will raise an error.

        :raises ValueError: if the inputs are not compatible with phonopy standards
        """
        self._if_can_modify()
        ph = self.get_phonopy_instance()

        try:
            ph.generate_displacements(
                distance=distance,
                is_plusminus=is_plusminus,
                is_diagonal=is_diagonal,
                is_trigonal=is_trigonal,
                number_of_snapshots=number_of_snapshots,
                random_seed=random_seed,
                temperature=temperature,
                cutoff_frequency=cutoff_frequency,
            )
        except ValueError as err:
            raise ValueError('one or more input types are not accepted') from err

        self.base.attributes.set('displacement_dataset', copy.deepcopy(ph.dataset))

    def set_displacements_from_dataset(self, dataset):
        """Set displacements for frozen phonon calculation from a dataset.
        Useful if you want to set displacements from a previously random generated
        displacement dataset, or for setting dataset for self-consistent harmonic
        approximation.

        :param dataset: dictionary or array like (numpy or list), compatible with phonopy

        :raises ValueError: if the inputs are not compatible with phonopy standards
        """
        self._if_can_modify()
        ph = self.get_phonopy_instance()

        if isinstance(dataset, dict):
            try:
                ph.dataset = dataset
            except RuntimeError as err:
                raise ValueError('non compatible format') from err
        elif isinstance(dataset, (np.ndarray, list)):
            try:
                ph.displacements = dataset
            except RuntimeError as err:
                raise ValueError('non compatible format') from err
        else:
            raise ValueError('type not accepted')

        self.base.attributes.set('displacement_dataset', copy.deepcopy(ph.dataset))

    def get_supercells_with_displacements(self):
        """Get the supercells with displacements for frozen phonon calculation.

        :returns: dictionary with StructureData nodes (note: not stored), None if
            the displacement dataset has not been set
        """
        ph = self.get_phonopy_instance(**{'distinguish': True})
        try:
            ph.dataset = self.displacement_dataset

            supercells = ph.supercells_with_displacements
            mapping = self.kinds_map
            structures_dict = {}

            # this will make the labels more nice to read - not really needed though
            digits = len(str(len(supercells)))

            for i, scell in enumerate(supercells):
                # start from 1 - better choice for gathering forces
                # MAYBE JUST NUMBERS???
                label = f'supercell_{str(i + 1).zfill(digits)}'
                structures_dict.update({label: phonopy_atoms_to_structure(scell, mapping)})

            return structures_dict
        except AttributeError:
            return None

    def generate_displacement_dataset(
        self,
        distance=0.01,
        is_plusminus='auto',
        is_diagonal=True,
        is_trigonal=False,
        number_of_snapshots=None,
        random_seed=None,
        temperature=None,
        cutoff_frequency=None,
    ):
        """Return the displacement dataset for frozen phonon calculation.
        (see `phonopy.Phonopy.generate_displacements` for a complete description of the inputs)

        :raises ValueError: if the inputs are not compatible with phonopy standards
        """
        ph = self.get_phonopy_instance()

        try:
            ph.generate_displacements(
                distance=distance,
                is_plusminus=is_plusminus,
                is_diagonal=is_diagonal,
                is_trigonal=is_trigonal,
                number_of_snapshots=number_of_snapshots,
                random_seed=random_seed,
                temperature=temperature,
                cutoff_frequency=cutoff_frequency,
            )
        except ValueError as err:
            raise ValueError('one or more inputs made the phonopy instance to raise an error') from err

        return copy.deepcopy(ph.dataset)

    @property
    def calcfunctions(self):
        """Namespace to access the calcfunction utilities."""
        from aiida_phonopy.calculations.functions.data_utils import CalcfunctionMixin

        return CalcfunctionMixin(data_node=self)

    @staticmethod
    def generate_preprocess_data(
        structure: orm.StructureData,
        displacement_generator=None,
        supercell_matrix=None,
        primitive_matrix=None,
        symprec=None,
        is_symmetry=None,
        distinguish_kinds=None,
    ):
        """Returns a complete stored PreProcessData node.

        :param structure: structure data node representing the unitcell
        :type structure: orm.StructureData
        :param displacement_generator: dictionary containing the info for generating the displacements,
            defaults to phonopy default (see phonopy doc)
        :type displacement_generator: orm.Dict
        :param supercell_matrix: supercell matrix, defaults to diag(1,1,1)
        :type supercell_matrix: orm.List, optional
        :param primitive_matrix: primitive matrix, defaults to "auto"
        :type primitive_matrix: orm.List or orm.List, optional
        :param symprec: symmetry precision on atoms, defaults to 1e-5
        :type symprec: orm.Float, optional
        :param is_symmetry: if using space group symmetry, defaults to True
        :type is_symmetry: orm.Bool, optional
        :param distinguish_kinds: if distinguish names of same specie by symmetry, defaults to True
        :type distinguish_kinds: orm.Bool, optional
        :return: PreProcessData node
        """
        from aiida_phonopy.calculations.functions.data_utils import generate_preprocess_data_from_distance

        kwargs = {}

        kwargs['structure'] = structure

        if displacement_generator is not None:
            kwargs['displacement_generator'] = displacement_generator

        if supercell_matrix is not None:
            kwargs['supercell_matrix'] = supercell_matrix

        if primitive_matrix is not None:
            kwargs['primitive_matrix'] = primitive_matrix

        if symprec is not None:
            kwargs['symprec'] = symprec

        if is_symmetry is not None:
            kwargs['is_symmetry'] = is_symmetry

        if distinguish_kinds is not None:
            kwargs['distinguish_kinds'] = distinguish_kinds

        return generate_preprocess_data_from_distance(**kwargs)



# -*- coding: utf-8 -*-
"""DataTypes for handling phonopy and frozen phonons calculations."""
from .force_constants import *
from .phonopy import *
from .preprocess import *
from .raw import *

__all__ = ('RawData', 'PreProcessData', 'PhonopyData', 'ForceConstantsData')



# -*- coding: utf-8 -*-
"""Utilities to deal with various mapping data structures."""


def get_logging_container():
    """Return an `AttributeDict` that can be used to map logging messages to certain log levels.

    This datastructure is useful to add log messages in a function that does not have access to the right logger. Once
    returned, the caller who does have access to the logger can then easily loop over the contents and pipe the messages
    through the actual logger.

    :return: :py:class:`~aiida.common.extendeddicts.AttributeDict`
    """
    from aiida.common import AttributeDict

    return AttributeDict({
        'debug': [],
        'info': [],
        'warning': [],
        'error': [],
        'critical': [],
    })


def _case_transform_dict(dictionary, dict_name, func_name, transform):
    from collections import Counter

    from aiida.common import InputValidationError

    if not isinstance(dictionary, dict):
        raise TypeError(f'{func_name} accepts only dictionaries as argument, got {type(dictionary)}')
    new_dict = dict((transform(str(k)), v) for k, v in dictionary.items())
    if len(new_dict) != len(dictionary):
        num_items = Counter(transform(str(k)) for k in dictionary.keys())
        double_keys = ','.join([k for k, v in num_items if v > 1])
        message = (
            f'Inside the dictionary `{dict_name}` there are the following keys that '
            f'are repeated more than once when compared case-insensitively: {double_keys}.'
            'This is not allowed.'
        )
        raise InputValidationError(message)
    return new_dict


def _lowercase_dict(dictionary, dict_name):
    return _case_transform_dict(dictionary, dict_name, '_lowercase_dict', str.lower)


def _uppercase_dict(dictionary, dict_name):
    return _case_transform_dict(dictionary, dict_name, '_uppercase_dict', str.upper)



# -*- coding: utf-8 -*-
"""Calcfunctions Utils for aiida-phonopy DataTypes."""
from aiida import orm
from aiida.engine import calcfunction
from aiida.plugins import DataFactory

PreProcessData = DataFactory('phonopy.preprocess')
PhonopyData = DataFactory('phonopy.phonopy')

__all__ = (
    'get_unitcell_from_distance', 'get_primitive_from_distance', 'get_supercell_from_distance',
    'get_supercells_with_displacements_from_distance', 'get_displacements_from_distance',
    'get_preprocess_with_new_displacements_from_distance', 'generate_preprocess_data_from_distance',
    'generate_phonopy_data_full', 'CalcfunctionMixin'
)


@calcfunction
def get_unitcell_from_distance(preprocess_data: PreProcessData):
    """Get the unitcell of a PreProcessData as a StructureData."""
    structure_data = preprocess_data.get_unitcell()
    return structure_data


@calcfunction
def get_primitive_from_distance(preprocess_data: PreProcessData):
    """Get the primitive cell of a PreProcessData as a StructureData."""
    structure_data = preprocess_data.get_primitive_cell()
    return structure_data


@calcfunction
def get_supercell_from_distance(preprocess_data: PreProcessData):
    """Get the supercell (pristine) of a PreProcessData as a StructureData."""
    structure_data = preprocess_data.get_supercell()
    return structure_data


@calcfunction
def get_supercells_with_displacements_from_distance(preprocess_data: PreProcessData):
    """Get the supercells with displacements of a PreProcessData as a StructureData."""
    structures_data = preprocess_data.get_supercells_with_displacements()
    return structures_data


@calcfunction
def get_displacements_from_distance(preprocess_data: PreProcessData):
    """Get the displacements of a PreProcessData as an ArrayData with array name `displacements`."""
    displacements = preprocess_data.get_displacements()
    the_displacements = orm.ArrayData()
    the_displacements.set_array('displacements', displacements)

    return the_displacements


@calcfunction
def generate_preprocess_data_from_distance(
    structure: orm.StructureData,
    displacement_generator=None,
    supercell_matrix=None,
    primitive_matrix=None,
    symprec=None,
    is_symmetry=None,
    distinguish_kinds=None,
):
    """Returns a complete stored PreProcessData node.

    :param structure: structure data node representing the unitcell
    :type structure: orm.StructureData
    :param displacement_generator: dictionary containing the info for generating the displacements
    :type displacement_generator: orm.Dict
    :param supercell_matrix: supercell matrix, defaults to diag(1,1,1)
    :type supercell_matrix: orm.List, optional
    :param primitive_matrix: primitive matrix, defaults to "auto"
    :type primitive_matrix: orm.List or orm.List, optional
    :param symprec: symmetry precision on atoms, defaults to 1e-5
    :type symprec: orm.Float, optional
    :param is_symmetry: if using space group symmetry, defaults to True
    :type is_symmetry: orm.Bool, optional
    :param distinguish_kinds: if distinguish names of same specie by symmetry, defaults to True
    :type distinguish_kinds: orm.Bool, optional

    :return: PreProcessData node
    """
    kwargs = {}

    kwargs['structure'] = structure

    if displacement_generator is not None:
        displ_generator = displacement_generator.get_dict()
    else:
        displ_generator = {
            'distance': 0.01,
            'is_plusminus': 'auto',
            'is_diagonal': True,
            'is_trigonal': False,
            'number_of_snapshots': None,
            'random_seed': None,
            'temperature': None,
            'cutoff_frequency': None,
        }

    if primitive_matrix is not None:
        if isinstance(primitive_matrix, orm.List):
            kwargs['primitive_matrix'] = primitive_matrix.get_list()
        if isinstance(primitive_matrix, orm.Str):
            kwargs['primitive_matrix'] = primitive_matrix.value
    else:
        kwargs['primitive_matrix'] = 'auto'

    if supercell_matrix is not None:
        kwargs['supercell_matrix'] = supercell_matrix.get_list()

    if symprec is not None:
        kwargs['symprec'] = symprec.value

    if is_symmetry is not None:
        kwargs['is_symmetry'] = is_symmetry.value

    if distinguish_kinds is not None:
        kwargs['distinguish_kinds'] = distinguish_kinds.value

    preprocess = PreProcessData(**kwargs)

    preprocess.set_displacements(**displ_generator)

    return preprocess


@calcfunction
def get_preprocess_with_new_displacements_from_distance(
    preprocess_data: PreProcessData, displacement_generator: orm.Dict
):
    """Get a new PreProcessData from an old one from new displacement generator settings."""
    displacement_dataset = preprocess_data.generate_displacement_dataset(**displacement_generator.get_dict())

    preprocess = PreProcessData(
        structure=preprocess_data.get_unitcell(),
        supercell_matrix=preprocess_data.supercell_matrix,
        primitive_matrix=preprocess_data.primitive_matrix,
        symprec=preprocess_data.symprec,
        is_symmetry=preprocess_data.is_symmetry,
        distinguish_kinds=preprocess_data.distinguish_kinds,
    )

    preprocess.set_displacements_from_dataset(displacement_dataset)

    return preprocess


@calcfunction
def generate_phonopy_data_full(preprocess_data: PreProcessData, nac_parameters=None, **forces_dict):
    """Create a PhonopyData node from a PreProcess(Phonopy)Data node, storing forces and (optionally)
        non-analytical constants.

        `Forces` must be passed as **kwargs**, since we are calling a calcfunction with a variable
        number of supercells forces.

        :param nac_parameters: ArrayData containing 'dielectric' and 'born_charges' as arrays
            with their correct shape
        :param forces_dict: dictionary of supercells forces as ArrayData stored as `forces`, each Data
            labelled in the dictionary in the format `{prefix}_{suffix}`.
            The prefix is common and the suffix corresponds to the suffix number of the supercell with
            displacement label given from the `get_supercells_with_displacements` method.

            For example:
                {'forces_1':ArrayData, 'forces_2':ArrayData}
                <==>
                {'supercell_1':StructureData, 'supercell_2':StructureData}
                and forces in each ArrayData stored as 'forces',
                i.e. ArrayData.get_array('forces') must not raise error

            .. note: if residual forces would be stored, label it with 0 as suffix.
    """
    import numpy as np

    # Getting the prefix
    for key in forces_dict:
        try:
            int(key.split('_')[-1])
            # dumb way of getting the prefix including cases of multiple `_`, e.g. `force_calculation_001`
            prefix = key[:-(len(key.split('_')[-1]) + 1)]
        except ValueError as err:
            raise ValueError(f'{key} is not an acceptable key, must finish with an int number.') from err

    forces_0 = forces_dict.pop(f'{prefix}_0', None)

    # Setting force sets array
    force_sets = [0 for _ in forces_dict]

    # Filling arrays in numeric order determined by the key label
    for key, value in forces_dict.items():
        index = int(key.split('_')[-1])
        force_sets[index - 1] = value.get_array('forces')

    # Finilizing force sets array
    sets_of_forces = np.array(force_sets)

    # Setting data on a new PhonopyData
    new_phonopy_data = PhonopyData(preprocess_data=preprocess_data)
    new_phonopy_data.set_forces(sets_of_forces=sets_of_forces)

    if forces_0 is not None:
        new_phonopy_data.set_residual_forces(forces=forces_0.get_array('forces'))

    if nac_parameters is not None:
        new_phonopy_data.set_dielectric(nac_parameters.get_array('dielectric'))
        new_phonopy_data.set_born_charges(nac_parameters.get_array('born_charges'))

    return new_phonopy_data


class CalcfunctionMixin:
    """Set of calcfunctions to be called from the aiida-phonopy DataTypes."""

    def __init__(self, data_node):
        self._data_node = data_node

    def get_unitcell(self):
        """Get the unitcell as a StructureData through a calfunction."""
        return get_unitcell_from_distance(preprocess_data=self._data_node)

    def get_primitive_cell(self):
        """Get the primitive cell as a StructureData through a calfunction."""
        return get_primitive_from_distance(preprocess_data=self._data_node)

    def get_supercell(self):
        """Get the supercell (pristine) as a StructureData through a calfunction."""
        return get_supercell_from_distance(preprocess_data=self._data_node)

    def get_supercells_with_displacements(self):
        """Get the supercells with displacements as a StructureData through a calfunction."""
        return get_supercells_with_displacements_from_distance(preprocess_data=self._data_node)

    def get_displacements(self):
        """Get the displacements as an ArrayData through a calfunction."""
        return get_displacements_from_distance(preprocess_data=self._data_node)

    def get_preprocess_with_new_displacements(self, displacement_generator: orm.Dict):
        """Create a PreProcessData node from a PreProcess/PhonopyData with a new set of displacements.

        :param displacement_generator: a `storable` dictionary  """
        return get_preprocess_with_new_displacements_from_distance(
            preprocess_data=self._data_node, displacement_generator=displacement_generator
        )

    def generate_full_phonopy_data(self, nac_parameters=None, **forces_dict):
        """Create a PhonopyData node from a PreProcess(Phonopy)Data node, storing forces and (optionally)
        non-analytical constants.

        `Forces` must be passed as **kwargs**, since we are calling a calcfunction with a variable
        number of supercells forces.

        :param nac_parameters: ArrayData containing 'dielectric' and 'born_charges' as arrays
            with their correct shape
        :param forces_dict: dictionary of supercells forces as ArrayData stored as `forces`, each Data
            labelled in the dictionary in the format `{prefix}_{suffix}`.
            The prefix is common and the suffix corresponds to the suffix number of the supercell with
            displacement label given from the `get_supercells_with_displacements` method

            For example:
                {'forces_1':ArrayData, 'forces_2':ArrayData} goes along with
                {'supercell_1':StructureData, 'supercell_2':StructureData}
                and forces in each ArrayData stored as 'forces',
                i.e. ArrayData.get_array('forces') must not raise error
        """
        return generate_phonopy_data_full(preprocess_data=self._data_node, nac_parameters=nac_parameters, **forces_dict)



# -*- coding: utf-8 -*-
"""Utilities for CalcJob resources."""


def get_default_options(max_num_machines=1, max_wallclock_seconds=300, with_mpi=False):
    """Return an instance of the options dictionary with the minimally required parameters for a `CalcJob`.

    :param max_num_machines: set the number of nodes, default=1
    :param max_wallclock_seconds: set the maximum number of wallclock seconds, default=1800
    :param with_mpi: whether to run the calculation with MPI enabled
    """
    return {
        'resources': {
            'num_machines': int(max_num_machines)
        },
        'max_wallclock_seconds': int(max_wallclock_seconds),
        'withmpi': with_mpi,
    }



# -*- coding: utf-8 -*-
"""
This module defines the class for force constants data.
"""

from aiida.orm import ArrayData
import numpy as np

from .raw import RawData


class ForceConstantsData(RawData, ArrayData):  # pylint: disable=too-many-ancestors
    """
    This class provides the force constants data, along with the non-analytical constants
    (optional) and the structure information (unitcell, supercell, ...).
    """

    def __init__(
        self,
        structure=None,
        phonopy_atoms=None,
        supercell_matrix=None,
        primitive_matrix=None,
        symprec=1e-05,
        is_symmetry=True,
        **kwargs
    ):
        kwargs['structure'] = structure
        kwargs['phonopy_atoms'] = phonopy_atoms
        kwargs['supercell_matrix'] = supercell_matrix
        kwargs['primitive_matrix'] = primitive_matrix
        kwargs['symprec'] = symprec
        kwargs['is_symmetry'] = is_symmetry

        super().__init__(**kwargs)

    def get_phonopy_instance(self, **kwargs):
        """Return a `phonopy.Phonopy` object with force and nac parameters (if set).

        :param kwargs: compatible keys with the super method in `RawData`
        """
        ph_instance = super().get_phonopy_instance(**kwargs)

        if self.displacement_dataset is not None:
            ph_instance.dataset = self.displacement_dataset

        if self.force_constants is not None:
            ph_instance.set_force_constants(self.force_constants)

        return ph_instance

    @property
    def force_constants(self):
        """Get force force constants matrix."""
        try:
            the_forces = self.get_array('force_constants')
        except KeyError:
            the_forces = None
        return the_forces

    def set_force_constants(self, force_constants):
        """Set force constants matrix.

        :param force_constants: array of force constants matrix in compact or full format
        :param type: (n_patom = atoms in primitive cell, n_satom =  atoms in supercell)
            * Compact format: (n_patom, n_satom, 3, 3)
            * Full format: (n_satom, n_satom, 3, 3)

        :raises:
            * TypeError: if the format is not of the correct type
            * ValueError: if the format is not compatible
            * RuntimeError: if the displacement dataset was not initialize in input
        """
        self._if_can_modify()

        if not isinstance(force_constants, (list, np.ndarray)):
            raise TypeError('the input is not of the correct type')

        n_satoms = len(self.get_supercell().sites)
        n_patoms = len(self.get_primitive_cell().sites)

        fc = np.array(force_constants)

        if fc.shape in [(n_patoms, n_satoms, 3, 3), (n_satoms, n_satoms, 3, 3)]:
            self.set_array('force_constants', fc)
        else:
            raise ValueError('the array is not of the correct shape')



# -*- coding: utf-8 -*-
"""Abstract workflow for automatic frozen phonons calculations."""

from abc import ABCMeta

from aiida import orm
from aiida.engine import WorkChain

from aiida_phonopy.data import ForceConstantsData, PhonopyData, PreProcessData


def validate_matrix(value, _):
    """Validate the `supercell_matrix` and `primitive_matrix` inputs."""
    import numpy as np

    if not isinstance(value, (list, orm.List, np.ndarray)):
        return 'value is not of the right type; only `list`, `aiida.orm.List` and `numpy.ndarray`'

    if isinstance(value, np.ndarray):
        value = value.tolist()

    if not len(value) == 3:
        return 'need exactly 3 diagonal elements or 3x3 arrays.'

    for row in value:
        if isinstance(row, list):
            if not len(row) in [0, 3]:
                return 'matrix need to have 3x1 or 3x3 shape.'
            for element in row:
                if not isinstance(element, (int, float)):
                    return (
                        f'type `{type(element)}` of {element} is not an accepted '
                        'type in matrix; only `int` and `float` are valid.'
                    )


def validate_positive_integer(value, _):
    """Validate that `value` is positive."""
    if not value.value > 0:
        return f'{value} is not positive.'


def validate_nac(value, _):
    """Validate that `value` is positive."""
    try:
        value.get_array('dielectric')
        value.get_array('born_charges')
    except KeyError:
        return 'data does not contain `dieletric` and/or `born_charges` arraynames.'


def validate_inputs(inputs, _):
    """Validate the entire inputs namespace."""
    ensamble_inputs = ['structure', 'supercell_matrix', 'primitive_matrix', 'symprec', 'is_symmetry']

    given_inputs = []

    for input_ in ensamble_inputs:
        if input_ in inputs:
            given_inputs.append(input_)

    if 'preprocess_data' in inputs and given_inputs:
        return 'too many inputs have been provided.'

    if given_inputs and 'structure' not in given_inputs:
        return 'a structure data is required'

    if not given_inputs and not 'preprocess_data' in inputs:
        return 'at least one between `preprocess_data` and `structure` must be provided in input'


class PhonopyWorkChain(WorkChain, metaclass=ABCMeta):
    """
    Abstract Workflow to compute automatically the force sets and force constants
    of a given structure using the frozen phonons approach. Phonopy is used to produce
    structures with displacements, while the forces are calculated with a quantum engine of choice.

    This workchain is meant to be used as a base for other specific force calculato plugin workchains,
    or as an example on how to set a possible workchain/workflow. For this reason, the outline of
    this class is not defined, while it provides the inputs and a `setup` method, which can be used
    in a specific workflow outline. Ideally, the workflow would look like:

    1. Setup the preprocess data.

        This is already provided in this class. It setups a `PreProcessData` node, from where
        supercell, primitive cell and supercells with displacements can be easily extracted using
        the methods of the nodes. This node can be taken from `self.ctx.preprocess_data`, and used
        during the outline of the workflow.

    2. Run supercells using the selected quantum engine/force calculator code.

        In specific code implementations, a force calculation on supercells needs to be run.
        To get these supercells, one need simply to run:

        ```self.ctx.preprocess_data.calcfunctions.get_supercells_with_displacements()```

        This will return a dictionary with all the supercells as StructureData to run for the phonon calculation.
        The keys of this dictionary are of the type `supercell_{number}`, where `number` is an integer.
        These numbers are essentials since the `phonopy` force sets is generated following these numbers,
        in order to make sure to refer to the correct displacement. Thus, it is required to keep track
        of them.
        Moreover,a calculation over the pristine supercell structure should be run before hand as reference.
        This structure can instead be gotten via:

        ```self.ctx.preprocess_data.calcfunctions.get_supercell()```

        This will return a StructureData without any label.

        For an example of implementation, refer to aiidateam/aiida-common-worfklows.

        * Note: some type of force calculation needs to map some variables from the unitcell to the supercell
        (and in certain case even the primitive cell), e.g. the atomic spin in VASP. Since this is code dependent,
        you will need to map these parameters before launching the force calculation of a certain supercell
        with displacement. This information can be gotten via:

        ```self.ctx.preprocess_data.get_cells_mappings()```

        Moreover, consider that cells in phonopy will always (re)fold the atoms in order to have positive coordinates.

    3. Inspect all runs and expose the forces and energies (not mandatory) outputs.

        * Suggested: when the calculation on each supercell has finished (correctly)
        expose the output forces (and energies) in the dynamical `supercells_forces(energies)` namespace(s).
        Provide each supercell forces as an `ArrayData` with the forces stored as `forces`
        (e.g. if your code plugin stores  the forces in `TrajectoryData`, extract them with a `calcfunction`).
        Expose each `ArrayData` choosing a **common prefix**, while as **suffix use
        _{number}**, with `{number}` referring to the correspective supercell label suffix (that you are supposed to
        keep track somewhere, e.g. in the label of the code calculation/workchain).
        Now you can gather all the information in one data noe, i.e. in a `PhonopyData` node.
        To do so, you can simple run:

        ```self.ctx.preprocess_data.calcfunctions.generate_full_phonopy_data(**self.outputs.supercells_forces)```

        and then expose it as output in the `output_phonopy_data` namespace.

        * Alternatively: instead of exposing the supercell forces as outputs, you can directly gather all the forces
        in a dictionary and run directly to the `generate_full_phonopy_data` method using this dictionary (always using
        the double *).

        See the implementation in aiidateam/aiida-common-workflows for an example.

    4. (optional) Run the non-analytical constants on the primitive cell.

        Non-analytical constants should be run for polar insulators. These require usually a linear response code
        or a finite difference approach (e.g. using the electric enthalpy). Since this is usually the most expensive
        part, you should run them on the primitive cell. To get it, use:

        ```self.ctx.preprocess_data.calcfunctions.get_primitive_cell()```

        If you compute also these, collect the dielectric tensor and the effectic born charges in an ArrayData,
        with the arraynames `dielectric` and `born_charges` (in Cartesian coordinates!).
        Then, gather all the information of nac and forces in a unique `PhonopyData` via:

        ```
        self.ctx.preprocess_data.calcfunctions.generate_full_phonopy_data(
            nac_parameters=nac_paramters,
            **self.outputs.supercells_forces
            )
        ```

        and expose the output.

        * Note: we require in the input for generating the full phonopy data, to give the nac in the primitive cell.
        The primitive cell of phonopy will just rotate the lattice vectors, thus mantaining the Cartasian coordinate
        system. It can happen, though, that the unitcell is not the primitive cell of the system, meaning that the
        primitive cell will contain less atoms. We expect in input the nac computed on this number of atoms. If you
        want, for some reason, compute the nac on the unitcell, you will need to get the reduced nac.
        To do so, you can consider using a built-in function in phonopy, namely:

        ```phonopy.structure.symmetry.elaborate_borns_and_epsilon```

    """

    _ENABLED_DISPLACEMENT_GENERATOR_FLAGS = {
        'distance': [float],
        'is_plusminus': ['auto', float],
        'is_diagonal': [bool],
        'is_trigonal': [bool],
        'number_of_snapshots': [int, None],
        'random_seed': [int, None],
        'temperature': [float, None],
        'cutoff_frequency': [float, None],
    }

    _ENABLED_FC_OPTIONS_FLAGS = {
        'calculate_full_force_constants': [bool],
        'fc_calculator': [str, None],
        'fc_calculator_options': [None],
    }

    @classmethod
    def define(cls, spec):
        """Define inputs, outputs, and outline."""
        super().define(spec)

        spec.input(
            'preprocess_data',
            valid_type=(PhonopyData, PreProcessData),
            required=False,
            help='The preprocess data for frozen phonon calcualtion.'
        )
        spec.input(
            'structure', valid_type=orm.StructureData, required=False, help='The structure at equilibrium volume.'
        )
        spec.input(
            'supercell_matrix',
            valid_type=orm.List,
            required=False,
            validator=validate_matrix,
            help=(
                'The matrix used to generate the supercell from the input '
                'structure in the List format. Allowed shapes are 3x1 and 3x3 lists.'
            ),
        )
        spec.input(
            'primitive_matrix',
            valid_type=orm.List,
            required=False,
            validator=validate_matrix,
            help=(
                'The matrix used to generate the primitive cell from the input '
                'structure in the List format. Allowed shapes are 3x1 and 3x3 lists.'
            ),
        )
        spec.input(
            'symmetry_tolerance',
            valid_type=orm.Float,
            required=False,
            help='Symmetry tolerance for space group analysis on the input structure.',
        )
        spec.input(
            'is_symmetry',
            valid_type=orm.Bool,
            required=False,
            help='Whether using or not the space group symmetries.',
        )
        spec.input(
            'displacement_generator',
            valid_type=orm.Dict,
            required=False,
            validator=cls._validate_displacements,
            help=(
                'Info for displacements generation. The following flags are allowed:\n ' +
                '\n '.join(f'{flag_name}' for flag_name in cls._ENABLED_DISPLACEMENT_GENERATOR_FLAGS)
            ),
        )
        spec.input(
            'fc_options',
            valid_type=orm.Dict,
            required=False,
            validator=cls._validate_fc_options,
            help=(
                'Options for force constants calculation (optional). The following flags are allowed:\n ' +
                '\n '.join(f'{flag_name}' for flag_name in cls._ENABLED_FC_OPTIONS_FLAGS)
            ),
        )
        spec.input(
            'nac_parameters',
            valid_type=orm.ArrayData,
            required=False,
            validator=validate_nac,
            help='Non-analytical parameters.',
        )
        spec.input_namespace(
            'options',
            help='Options for how to run the workflow.',
        )
        spec.input(
            'options.run_parallel_nac_forces',
            valid_type=bool,
            required=False,
            non_db=True,
            default=True,
            help='Whether running nac parameters and forces calculations in parallel.',
        )
        spec.inputs.validator = validate_inputs

        spec.output_namespace(
            'supercells',
            valid_type=orm.StructureData,
            dynamic=True,
            required=False,
            help='The supercells with displacements.'
        )
        spec.output_namespace(
            'supercells_forces',
            valid_type=orm.ArrayData,
            required=True,
            help='The forces acting on the atoms of each supercell.'
        )
        spec.output_namespace(
            'supercells_energies', valid_type=orm.Float, required=False, help='The total energy of each supercell.'
        )
        spec.output(
            'output_phonopy_data',
            valid_type=PhonopyData,
            help=(
                'The phonopy data with supercells displacements, forces and (optionally)'
                'nac parameters to use in the post-processing calculation.'
            )
        )
        spec.output(
            'output_force_constants',
            valid_type=ForceConstantsData,
            required=False,
            help='The matrix of force constants computed with finite displacements.'
        )

    @classmethod
    def _validate_displacements(cls, value, _):
        """Validate the ``displacements`` input namespace."""
        if value:
            value_dict = value.get_dict()
            enabled_dict = cls._ENABLED_DISPLACEMENT_GENERATOR_FLAGS
            unknown_flags = set(value_dict.keys()) - set(enabled_dict.keys())
            if unknown_flags:
                return (
                    f"Unknown flags in 'displacements': {unknown_flags}, "
                    f'allowed flags are {cls._ENABLED_DISPLACEMENT_GENERATOR_FLAGS.keys()}.'
                )
            invalid_values = [
                value_dict[key]
                for key in value_dict.keys()
                if not (type(value_dict[key]) in enabled_dict[key] or value_dict[key] in enabled_dict[key])
            ]
            if invalid_values:
                return f'Displacement options must be of the correct type; got invalid values {invalid_values}.'

    @classmethod
    def _validate_fc_options(cls, value, _):
        """Validate the ``fc_options`` input namespace."""
        if value:
            value_dict = value.get_dict()
            enabled_dict = cls._ENABLED_FC_OPTIONS_FLAGS
            unknown_flags = set(value_dict.keys()) - set(enabled_dict.keys())
            if unknown_flags:
                return (
                    f"Unknown flags in 'fc_options': {unknown_flags}, "
                    f'allowed flags are {cls._ENABLED_FC_OPTIONS_FLAGS.keys()}.'
                )
            invalid_values = [
                value_dict[key]
                for key in value_dict.keys()
                if not (type(value_dict[key]) in enabled_dict[key] or value_dict[key] in enabled_dict[key])
            ]
            if invalid_values:
                return f'Displacement options must be of the correct type; got invalid values {invalid_values}.'

    def setup(self):
        """Setup the workflow generating the PreProcessData."""
        if 'preprocess_data' in self.inputs:
            preprocess = self.inputs.preprocess_data
            if 'displacement_generator' in self.inputs:
                preprocess = preprocess.calcfunctions.get_preprocess_with_new_displacements(
                    self.inputs.displacement_generator
                )
        else:
            preprocess_inputs = {}
            for input_ in [
                'structure', 'supercell_matrix', 'primitive_matrix', 'symprec', 'is_symmetry', 'displacement_generator'
            ]:
                if input_ in self.inputs:
                    preprocess_inputs.update({input: self.inputs[input_]})
            preprocess = PreProcessData.generate_preprocess_data(**preprocess_inputs)

        self.ctx.preprocess_data = preprocess



# -*- coding: utf-8 -*-
"""
This module defines the class which wraps the :func:`phonopy.Phonopy` main class.
"""

import numpy as np

from .preprocess import PreProcessData


class PhonopyData(PreProcessData):  # pylint: disable=too-many-ancestors
    """
    This class wraps the :class:`phonopy.Phonopy` class. It represents the final Data node status
    of a frozen phonon calculaiton. It stores information regarding the pre-processing,
    the displacements and forces dataset, and the (eventual) non-analytical constants.

    .. note: direct calculation of properties from this class is still not implemented.
        Use :class:`~aiida_phonopy.calculations.phonopy.PhonopyCalculation` for
        post-processing this data node, keeping the provenance and having the
        data produced in the correct data type.
    """

    def __init__(self, preprocess_data: PreProcessData, **kwargs):

        if isinstance(preprocess_data, PreProcessData):
            kwargs['structure'] = preprocess_data.get_unitcell()
            kwargs['supercell_matrix'] = preprocess_data.supercell_matrix
            kwargs['primitive_matrix'] = preprocess_data.primitive_matrix
            kwargs['symprec'] = preprocess_data.symprec
            kwargs['is_symmetry'] = preprocess_data.is_symmetry
            kwargs['distinguish_kinds'] = preprocess_data.distinguish_kinds
            super().__init__(**kwargs)

            super().set_displacements_from_dataset(preprocess_data.displacement_dataset)

        else:
            raise ValueError('`preprocess_data` is not of the correct type')

    def set_displacements(self):
        raise RuntimeError('`displacements` cannot be changed for this Data')

    def set_displacements_from_dataset(self):
        raise RuntimeError('`displacements` cannot be changed for this Data')

    def get_phonopy_instance(self, subtract_residual_forces=None, **kwargs):
        """Return a `phonopy.Phonopy` object with force and nac parameters (if set).

        :param subtract_residual_forces: whether or not subract residual forces (if set)
        :type subtract_residual_forces: bool, defaults to False
        :param kwargs: compatible keys with the super method in `RawData`
        """
        if not isinstance(subtract_residual_forces, bool) and subtract_residual_forces is not None:
            raise TypeError('`subtract_residual_forces` not of the right type')

        if subtract_residual_forces and self.residual_forces is None:
            raise ValueError('`subtract_residual_forces` is set to True, but `residual_forces` are None')

        ph_instance = super().get_phonopy_instance(**kwargs)

        if self.displacement_dataset is not None:
            ph_instance.dataset = self.displacement_dataset

        if self.forces is not None:
            the_forces = self.forces
            if subtract_residual_forces:
                the_forces = the_forces - self.residual_forces
            ph_instance.forces = the_forces

        return ph_instance

    @property
    def residual_forces(self):
        """Get the residual forces calculated on the pristine (i.e. no displaced) supercell structure (if set)."""
        try:
            the_forces = self.get_array('residual_forces')
        except KeyError:
            the_forces = None
        return the_forces

    def set_residual_forces(self, forces):
        """Set the residual forces of the pristine supercell.

        :param forces: (atoms in supercell, 3) array shape

        :raises:
            * TypeError: if the format is not of the correct type
            * ValueError: if the format is not compatible
        """
        self._if_can_modify()

        if not isinstance(forces, (list, np.ndarray)):
            raise TypeError('the input is not of the correct type')

        natoms = len(self.get_supercell().sites)

        the_forces = np.array(forces)

        if the_forces.shape == (natoms, 3):
            self.set_array('residual_forces', the_forces)
        else:
            raise ValueError('the array is not of the correct shape')

    @property
    def forces(self):
        """Get forces per each supercell with displacements in the dataset as a unique array."""
        try:
            the_forces = self.get_array('forces')
        except KeyError:
            the_forces = None
        return the_forces

    def set_forces(self, sets_of_forces):
        """Set forces per each supercell with displacement in the dataset.

        :param sets_of_forces:  a set of atomic forces in displaced supercells. The order of
            displaced supercells has to match with that in displacement dataset.
        :param type: (supercells with displacements, atoms in supercell, 3) array shape

        :raises:
            * TypeError: if the format is not of the correct type
            * ValueError: if the format is not compatible
            * RuntimeError: if the displacement dataset was not initialize in input
        """
        self._if_can_modify()

        if not isinstance(sets_of_forces, (list, np.ndarray)):
            raise TypeError('the input is not of the correct type')

        if self.displacement_dataset is None:
            raise RuntimeError('the displacement dataset has not been set yet')

        nsupercells = len(self.displacements)
        natoms = len(self.get_supercell().sites)

        the_forces = np.array(sets_of_forces)

        if the_forces.shape == (nsupercells, natoms, 3):
            self.set_array('forces', the_forces)
        else:
            raise ValueError('the array is not of the correct shape')



# -*- coding: utf-8 -*-
"""Defines a `Parser` base class for `aiida-phonopy`.

All `Parser` implementations in `aiida-phonopy` must use this base class, not `aiida.parsers.Parser`.
"""
from aiida.parsers import Parser as _BaseParser

__all__ = ('Parser',)


class Parser(_BaseParser):  # pylint: disable=abstract-method
    """Custom `Parser` class for `aiida-phonopy` parser implementations."""

    def emit_logs(self, logging_dictionaries, ignore=None):
        """Emit the messages in one or multiple "log dictionaries" through the logger of the parser.

        A log dictionary is expected to have the following structure: each key must correspond to a log level of the
        python logging module, e.g. `error` or `warning` and its values must be a list of string messages. The method
        will loop over all log dictionaries and emit the messages it contains with the log level indicated by the key.

        Example log dictionary structure::

            logs = {
                'warning': ['Could not parse the `etot_threshold` variable from the stdout.'],
                'error': ['Self-consistency was not achieved']
            }

        :param logging_dictionaries: log dictionaries
        :param ignore: list of log messages to ignore
        """
        ignore = ignore or []

        if not isinstance(logging_dictionaries, (list, tuple)):
            logging_dictionaries = [logging_dictionaries]

        for logs in logging_dictionaries:
            for level, messages in logs.items():
                for message in messages:

                    if message is None:
                        continue

                    stripped = message.strip()

                    if not stripped or stripped in ignore:
                        continue

                    try:
                        getattr(self.logger, level)(stripped)
                    except AttributeError:
                        pass

    def exit(self, exit_code):
        """Log the exit message of the give exit code with level `ERROR` and return the exit code.

        This is a utility function if one wants to return from the parse method and automically add the exit message
        associated to the exit code as a log message to the node: e.g. `return self.exit(self.exit_codes.LABEL))`

        :param exit_code: an `ExitCode`
        :return: the exit code
        """
        self.logger.error(exit_code.message)
        return exit_code
